/*
d-lib.cpp
- Started 2019-12-18 by Eric Wallin
- No explicit project management, just compile this file.

================
!FIXME! / TODO / NOTE
- note: DON'T USE ISSPACE()! - crashes w/ bad input (use ch_is_white)
- note: DON'T USE STOL()! - crashes w/ bad input (use strtol w/ str_is_hex() check)
- note: DON'T USE getch()! - broken, replaced with read()
- note: do a cout flush before sleep!
================

Revision History:

2021-07-07 : 0x109
- can now enter filter frequencies directly & closest is picked
- updated to reflect changes in UI: Adel=>Wait, type 30=>0xf8

2021-06-30 : 0x108
- noise vmod now bump/dip @ -24dB (n_p9_ds type 0x84 => 0x4a)

2021-05-22 : 0x107
- interactive serial port selection, now check ports [0:31]
- added typical usage examples to help file

2021-05-04 : 0x106
- updated to reflect changes in UI

2021-04-29 : 0x105
- updated to reflect changes in SPI CSn & protected writes
- removed etok from ptoe command
- removed auto file list update for F2 (was blowing out)

2021-04-04 : 0x104
- updated to reflect changes

2021-03-25 : 0x103
- updated to reflect changes

2021-03-07 : 0x102
- serial port scan now automatically selects single open port
- had to make all const string "static" otherwise they get corrupted!  lordy...

2021-03-05 : 0x101
- NO more config files, ports 0:9 are scanned & selected at start
- ALL file reading now goes through file_to_str() in d-lib_pkg.cpp to nip any MS crap in the bud
- more refactoring in d-lib_preset.cpp
- combined console & serial files via ifdefs

2021-03-04 : 0x101
- open serial ports now listed @ port open failure
- new etok command, ptoe automatically calls it too
- lots of refactoring in d-lib_preset.cpp, some in main
- command renaming: 
- stopn => stop_num
- ptoes => ptoe_sys
- etops => etop_sys

2021-03-03
- "up" command updates old preset file to newest encoder scalings.

2021-02-25
- updated for new Adel knob

2021-02-23
- updated for new boot code location, preset locations, knobs

2021-01-11
- updated to reflect changes (encoder name & range changes)

2020-10-04
- updated to reflect changes (moved DISPLAY page)

2020-09-19
- F1 CTRL + (R, L) now +/-10 the hilited encoder value 

2020-09-16
- added live update from the F1 screen!
- changes to F1 editor hilite and +/-:
  CTRL + (UP, DN) +/-1 the hilited encoder value 
  CTRL + (R, L) +/-16 the hilited encoder value 
  ALT + (R, L, UP, DN) move hilite one UI screen
- added PRE_UPD_ADDR to d-lib_pkg.cpp
- removed F1 cla command.

2020-09-14
- mass renaming of commands ('f' for preset file is ambiguous at this point):
  ftoe => ptoe
  ftoe_sys => ptoes
  etof => etop
  etof_sys => etops
  stof => stop 
  ftos => ptos
- changed dump_usr command: <slot> <slot> stopn (slots to preset numbered files)
- numbered preset files are now all 3 char: ???.pst to better sort
- added SYS_ADDR & USR_ADDR to d-lib_pkg.cpp, removed from config files & process
- help file contents now read in @ start
- added safety timeouts to serial rx code
- spi_pump activity indicator speed up

2020-09-12
- added "dump_usr" command: slots [0:103] => numbered files
- added "ver" command to check SW version

2020-09-10
- fixed race condition between UART TX & RX (slot & knob value reads)
- now have separate RX UART burst & wait sleep times

2020-09-09
- updated MS compile (to Roger)

2020-08-20
- updated to reflect changes (new "Line" knob on "LEVELS" UI page)

2020-08-12
- updated to reflect changes (a few type changes)

2020-08-06
- broken with Linux Mint v20 update!
- issue was getch(), replaced with read, added cout flush before sleep
- other minor improvements to input code

2020-07-20
- moved serial port defines to d-lib_pkg.cpp, increased buffer size 4x
- using RX_MDLY for all polling

2020-07-19
- finally fixed Win32 delay by adding timeBeginPeriod(1), an issue which 
  has plagued me literally for years
- the issue with spi_pump not completing the reset on other prototype is due to 
  how differing FPGA versions (v1305 vs v1306) handle it (v1305 locks up)
- added progress indicator to btos, tweaked others

2020-07-17
- fixed PPst => PP knob name & type
- fixed oct type

2020-07-16
- slight edits to make mswin compile work

2020-07-14
- added function key designations to screen titles

2020-06-13
- updated to reflect changes (usr params shuffle)
- restored & edited "uf" command to update preset file param order

2020-06-12
- updated to reflect changes (sys params shuffle)

2020-06-06
- fixed hilite not tracking after delete of last file on F2 page

2020-06-03
- "cd" and "dir" alised to "chdir" and "ls"
- fixed auto-complete (was only working for word at end of cmd line)

2020-06-01
- added "cd" command to change working directory
- added "dir" command to show all files & directories
- removed file overwrite check for commands "stof "& "ltob"

2020-05-28
- updated to reflect changes
- new command "ktoe" which downloads current knob values to editor
- btos now goes in neg dir w/ neg start slot, limited to +/-99
- commented out "uf" update file command

2020-04-10
- fixed autocomplete: sv_str_match(sorted_i = false) 
- DEL key in files hilite mode now deletes file and exits mode
- auto switch to edit mode after ftoe and stoe
- command updf => uf (easier to type)
- simplified dir print to file for easier use later
- updated UI to reflect changes

2020-04-08
- new functions:
- pre_str_to_pre_uba()
- pre_uba_to_pre_str()
- dlp_update()
- new command: updf (update file)

2020-04-07
- updated to reflect changes

2020-04-03
- updated to reflect changes

2020-03-21
- updated to reflect osc changes

2020-03-20
- *.spi & *.bnk files now show up in F2 files view
  and can be deleted, etc.
- file extension optional

2020-03-19
- new command: <file> <slot> btos (bank file to slots)
- sw_pump => spi_pump (back to this)

2020-03-18
- new command: <file> ltob (list to bank file)
- ...and there was much renaming:
- ptof => print
- dtof => stof
- ftod => ftos
- dtoe => stoe
- etod => etos
- ftoes => ftoe_sys
- etofs => etof_sys
- dlist => stol
- spi_pump => sw_pump

2020-03-15
- added "etod" and "dtoe" commands.
- ui page rename: "CORRECT" => "EZ-PITCH"

2020-03-09
- Pcal & Vcal [-127:127] => [0:255] (type 0x43 => 0x83)

2020-03-07
- moved knobs on display, v_axis, p_axis ui pages

2020-03-06
- moved pv_fmod ui page, *_en knobs on it

2020-02-14
- lots of command renaming / elimination:
- rf, rsf, wf, wsf => ftoe, ftoes, etof, estof
- dlf, ulf => dtof, ftod
- dl => dlist
- ren => rene
- pf => ptof
- removed ed, owf, owsf, ul, dl to ed

2020-02-13
- changes to reflect recent D-Lev SW
- slots > 99 are system presets
- hilite now off by default
- type 0xff displays '0' value

2020-01-31
- FTDI chip speed-up code in d-lib_serial_linux.cpp

2020-01-30
- merged command result and error message systems

2020-01-23
- fixed sp_tx() in d-lib_serial_mswin.cpp (was crashing)
- works in win7 & win10; +1 to console width makes it work in xp

2020-01-22
- help file viewer sub in d-lib_file_view.cpp
- some function renaming in console files (for clarity)
- frittering around w/ mswin compile (XP is a nightmare)

2020-01-21
- fixed MSWin config file (no repeated backslashes necessary?)
- hilite now on by default on slots page

2020-01-20
- import from MSWin work on Win10 laptop
- implemented d-lib.cfg file for serial port config & 1st cmd
- new visual indication of pump & dl (dots)
- serial port open check before using it (prevent crashing)
- removed spi pump escape w/ kbhit (speedup)
- fixed isspace() in pkg
- everything that opens a file now closes it
- all int => int32_t

2020-01-16
- working on MSWin compile:
- new function us_sleep() in pkg.

2020-01-15
- new command "renf" to rename *.dlp file
- new command "copf" to copy *.dlp file
- ed hilite behaves more like hilite on other pages
- hilite mode default for dir F2
- fixed slots view init bug
- big cleanup of cmd decode
- some minor font changes

2020-01-14
- file list and slots pages can now accommodate zillions of entries
- new command "delf" to download slot directly to file (won't overwrite)
- separated ulf out from ul
- increased SLOT_ROWS 20 => 25
- commented out DEL key in file view (too dangerous?)

2020-01-13
- combined "dl" and "slots" commands
- can now edit from dl buffer (but not to)
- hilite works on slots screen, enter selects dl buffer to edit
- can now directly upload from file to slot via <file> <slot> ul

2020-01-12
- help file ~done
- new cla command to clear all preset enc values in ed mode
- new ID banner text on screens
- new pf file name display
- command rename: rmf => delf
- rewrote dir & bank display subs

2020-01-11
- bank & dir screens print to text file done
- converted serial code from obj to struct & subs
- preset struct now included in sub ports
- ui mode is now an enum
- command rename: del => rmf; prn => pf

2020-01-10
- new del command to delete *.dlp files
- del in dir mode deletes file @ hilite via new del command
- detailed cmd error messages
- fixed crashing on bad file read / isspace / stol
- more monkeying around with mode switching, ESC kills hilight & terminal modes

2020-01-09
- mass rename: ed_* => hl_*; ui_* => ed_*

2020-01-08
- bank view page working
- F1 UI, F2 dir, F3 bank, F4 term, w/ second press toggling edit mode for UI & dir
- preset object converted to struct & functions
- tracked down painful isspace() bug blowing out, now using ch_is_white() in pkg

2020-01-07
- UI edit hilite "warps" to screens w/ CTRL
- UI edit hilite inc/dec now via +/-
- file list page working ok

2020-01-06
- autocomplete working
- auto *.dlp extension
- worked on file list page a bit (now F1)
- terminal mode F4

2020-01-05
- have functions to do autocomplete

2020-01-04
- terminal mode accessable any time via F2
- file list (unordered) via F3
- files rename: lib* => d-lib*
- serial spi code neatening

2020-01-03
- can upload preset from UI and file
- can pump spi with software load
- added kbd_hit() to rx_wait()

2020-01-02
- can download preset to UI and file
- some subroutine shuffling / renaming

2019-12-30
- now opens serial port @ start, closes @ quit
- has optional terminal mode at start

2019-12-26
- now using CTRL rather than ALT in F1 edit mode

2019-12-25
- F1 toggles edit mode, ESC clears it:
  - up/dn/lt/rt: navigate
  - #: write param
  - # on title: write all params on that UI screen (use to zero all)
  - ALT up/dn: inc/dec param
  - ALT lt/rt: inc/dec param by 4

2019-12-23
- new owf and owsf commands to overwrite *.dlp files
- wf and wsf will not overwrite existing files
- in edit mode:
  - + or - alone will inc / dec param
  - whitespace alone will clear param
  - number followed by + or whitespace will entrer that number
  - number followed by - will enter that number as negative

2019-12-22
- UI screens draw working
- cmd line working
- preset file rd / wr / prn via cmd line working
- UI hilite & editing working
- MSWin compile option added 

2019-12-20
- preset file read and write working
- UI screens print to text file working
- can read & write UI knob locations
- manually constructed and saved my system preset

2019-12-19
- much basic preset functionality working
- can output preset UI screens to string

2019-12-18
- born

*/

// comment this out for POSIX compile
// #define _MSWIN

// c++ stuff
#include <bits/stdc++.h>
#include <stdint.h>
#ifdef _MSWIN
  #include <conio.h>
  #include <windows.h>
  #include <unistd.h>
#else
  #include <fcntl.h>
  #include <linux/serial.h>
  #include <sys/ioctl.h>
#endif

using namespace std;

// lib stuff
#include "d-lib_pkg.cpp"
#include "d-lib_console.cpp"
#include "d-lib_serial.cpp"
#include "d-lib_cmd.cpp"
#include "d-lib_preset.cpp"
#include "d-lib_dir.cpp"
#include "d-lib_term.cpp"
#include "d-lib_view_file.cpp"


//////////
// MAIN //
//////////

int main(int argc, char* argv[]) {

	// if there are command line arguments
	if (argc > 1) {  
		cout << "\nd-lib uses no command line arguements";
		return(-1);
	}

#ifdef _MSWIN
	// size the MSWIN console (width +1 for WinXP console?)
	con_size(CON_W+1, CON_H);
	static const string dev_base = "\\\\.\\COM";
	// for more accurate (~1ms) timing - NOTE: add -lwinmm to build!
	timeBeginPeriod(1);
#else
	static const string dev_base = "/dev/ttyUSB";  // breaks w/o "static"?
#endif

	// constants
	const int32_t FILE_Y = 26;
	const int32_t PREV_Y = 28;
	const int32_t CMD_Y = PREV_Y + 1;
	const int32_t CMD_W = CON_W;
	// file strings
	static const string help_fname = "d-lib_help.txt";
	string pre_fname = "";
	string prn_fname = "d-lib_print.txt";
	// command line
	const int32_t CMD_DEPTH = 32;  // cmd history depth
	bool cmd_f = false;  // cmd flag
	bool err_f = false;  // cmd error flag
	string cmd_str = "";  // cmd str
	string cmd_prev = "";  // cmd history
	string cmd_msg = "";  // cmd result message
	cmd_mem<CMD_DEPTH> cmd_hist;  // cmd buffer
	uint32_t cursor_pos = 0;
	// modes
	enum ui_modes { edit, files, slots } mode = edit; 
	bool first_f = true;  // do extra stuff @ start
	bool hl_f = false;  // hilite mode flag
	// coords & sycg
	int32_t ed_x = 0;  // edit coords
	int32_t ed_y = 0;
	int32_t dir_idx = 0;  // hilite edit idx
	int64_t key_i = 0;
	int32_t dl_min = 0;
	int32_t dl_max = 79;
	int32_t dl_idx = 0;
	// serial port
	sp_type sp;
	uint32_t dev_num = 0;
	const uint32_t dev_max = 31;
	bool dev_open[dev_max+1];  // +1 to hold all
	uint32_t devs = 0;
	uint32_t dev_last = 0;

	// output version
	cout << "-= D-Lev Preset Librarian (version 0x" << hex << LIBV << ") =-" << endl << endl; 
	cout.unsetf(ios::hex); // I hate c++
	
	// scan for open ports
	while (dev_num <= dev_max) {
		dev_open[dev_num] = sp_test_port(dev_base + to_string(dev_num));
		if (dev_open[dev_num]) { dev_last = dev_num; devs++;}
		dev_num++;
	}
	
	if (!devs) {
		cout << "No open ports!" << endl;
	}
	else {
		cout << "Open ports:" << endl;
		for (uint32_t i=0; i<=dev_max; i++) { if (dev_open[i]) { cout << " " << i; } }
	}

	// query for port
	dev_num = dev_last;
	cout << endl << "Select port (hit Enter for port " << dev_num << "): ";
	string dev_str;
	getline(cin, dev_str);
	if (str_is_hex(dev_str)) { dev_num = strtol(dev_str.c_str(), NULL, 0); }
	cout << endl;

	// open port & check
	sp_open_port(sp, dev_base + to_string(dev_num));
	cout << "sel  : " << dev_num << endl; 
	cout << "port : " << sp.dev << endl; 
	if (sp_open_err(sp)) {
		cout << "open : ERROR! (fd=" << sp.fd << ")" << endl;
	}
	else { 
		cout << "open : OK! (fd=" << sp.fd << ")" << endl;
		if (sp_config_port(sp)) { cout << "cfg  : ERROR!" << endl; }
		else { cout << "cfg  : OK! " << endl; }
	}

	// read help file
	if (view_file(help_fname, CON_H, true)) { 
		cout << "can't read help file:" << help_fname << endl << endl;
	}

#ifndef _MSWIN
	// switch POSIX kbd mode to non-blocking
	kbd_mode(true, false);
#endif

	// pause to display addresses and port status
	cout << endl << "press any key to continue..." << endl;
	kbd_wait();

	// directory listings
	vector<string> dir_all, dir_dlp;
	dir_proc(dir_all, dir_dlp, false);  // read, sort, filter by ext, remove ext

	// downloaded preset contents
	vector<string> dl_pre_sv;

	// slot names
	vector<string> dl_names_sv;

	// editor preset
	ed_pre_t ed_pre;

	// loop forever
	while (true) {

		if (!first_f) {  // skip once @ start

		    //////////////
		    // key proc //
		    //////////////

			// get key code (blocking)
			key_i = kbd_getkey();
			bool key_used = false;

			// in edit hilite mode
			if ((mode == edit) && hl_f && !key_used) {
				key_used = true;  // default
				int32_t dx = 0;
				int32_t dy = 0;
				// quit hilite mode
				if (key_i == KBD_ESC) { hl_f = false; }
				// hilite navigation
				else if (key_i == KBD_UP) { dy = -1; }
				else if (key_i == KBD_DN) { dy = 1; }
				else if (key_i == KBD_LT) { dx = -1; }
				else if (key_i == KBD_RT) { dx = 1; }
				else if (key_i == KBD_ALT_UP) { dy = -4; }
				else if (key_i == KBD_ALT_DN) { dy = 4; }
				else if (key_i == KBD_ALT_LT) { dx = -2; }
				else if (key_i == KBD_ALT_RT) { dx = 2; }
				else if (key_i == KBD_HOME) { ed_xy_home(ed_x, ed_y); }
				else if (key_i == KBD_END) { ed_xy_end(ed_x, ed_y); }
				// inc/dec
				else if (key_i == KBD_CTRL_UP) { ed_enc_delta(sp, ed_pre, 1, ed_x, ed_y, true); }
				else if (key_i == KBD_CTRL_DN) { ed_enc_delta(sp, ed_pre, -1, ed_x, ed_y, true); }
				else if (key_i == KBD_CTRL_RT) { ed_enc_delta(sp, ed_pre, 10, ed_x, ed_y, true); }
				else if (key_i == KBD_CTRL_LT) { ed_enc_delta(sp, ed_pre, -10, ed_x, ed_y, true); }
				// zero entry
				else if (key_i == KBD_DEL) { cmd_str = "0 "; }
				// default
				else { key_used = false; }
				// recalc idx
				ed_xy_move(ed_x, ed_y, dx, dy);
			}
			// in files hilite mode
			if ((mode == files) && hl_f && !key_used) {
				key_used = true;  // default
				int32_t dx = 0;
				int32_t dy = 0;
				// quit hilite mode
				if (key_i == KBD_ESC) { hl_f = false; }
				// hilite navigation
				else if (key_i == KBD_UP) { dy = -1; }
				else if (key_i == KBD_DN) { dy = 1; }
				else if (key_i == KBD_LT) { dx = -1; }
				else if (key_i == KBD_RT) { dx = 1; }
				else if (key_i == KBD_HOME) { dir_idx = 0; }
				else if (key_i == KBD_END) { dir_idx = dir_all.size()-1; }
				// delete file & exit hilite mode
				else if (key_i == KBD_DEL) { 
					cmd_str = dir_all[dir_idx] + " delf ";
					cursor_pos = cmd_str.size();
					hl_f = false;
				}
				// copy hilited to cmd line & exit hilite mode
				else if (key_i == KBD_ENT) {
					cmd_str += dir_all[dir_idx];
					cursor_pos = cmd_str.size();
					hl_f = false;
				}
				else { key_used = false; }
				// recalc idx
				dir_idx = xy_move(dir_idx, dx, dy, dir_all.size()-1, DIR_ROWS);
			}
			// in slots hilite mode
			if ((mode == slots) && hl_f && !key_used) {
				key_used = true;  // default
				int32_t dx = 0;
				int32_t dy = 0;
				// quit hilite mode
				if (key_i == KBD_ESC) { hl_f = false; }
				// hilite navigation
				else if (key_i == KBD_UP) { dy = -1; }
				else if (key_i == KBD_DN) { dy = 1; }
				else if (key_i == KBD_LT) { dx = -1; }
				else if (key_i == KBD_RT) { dx = 1; }
				else if (key_i == KBD_HOME) { dl_idx = 0; }
				else if (key_i == KBD_END) { dl_idx = dl_pre_sv.size()-1; }
				// copy hilited to cmd line & exit hilite mode
				else if (key_i == KBD_ENT) { 
					cmd_str += to_string(dl_idx + dl_min); 
					cursor_pos = cmd_str.size();
					hl_f = false;
				}
				else { key_used = false; }
				// recalc idx
				dl_idx = xy_move(dl_idx, dx, dy, max(0, int32_t(dl_pre_sv.size()-1)), SLOTS_ROWS);
			}
			// command line ops
			if (!key_used && !cmd_edit(cmd_str, cursor_pos, key_i)) {
				// edit mode
				if (key_i == KBD_F1) {
					if (mode != edit) { 
						mode = edit;
						hl_f = false; 
						con_cls();
					}
					else if (!hl_f) { 
						hl_f = !hl_f; 
						cmd_str.clear();
						cursor_pos = cmd_str.size();
					}
					else { hl_f = false; }
				}
				// directory mode
				else if (key_i == KBD_F2) {
					if (mode != files) { 
						mode = files;
						// dir_proc(dir_all, dir_dlp, false);  // update file list
						dir_idx = xy_move(dir_idx, 0, 0, dir_all.size()-1, DIR_ROWS);
						hl_f = false; 
						con_cls();
					}
					else { hl_f = !hl_f; }
				}
				// slots mode
				else if (key_i == KBD_F3) {
					if (mode != slots) { 
						mode = slots;
						hl_f = false; 
						con_cls(); 
					}
					else { hl_f = !hl_f; }
				}
				// cmd_hist recall
				else if (key_i == KBD_UP) {
					if (cmd_hist.rd(cmd_str, true)) { cursor_pos = cmd_str.size(); }
				}
				else if (key_i == KBD_DN) {
					if (cmd_hist.rd(cmd_str, false)) { cursor_pos = cmd_str.size(); }
				}
				// file name autocomplete
				else if (key_i == KBD_TAB) {
					cmd_acomp(cmd_str, cursor_pos, dir_all); 
				}
			}
		}
		

		////////////////
		// token proc //
		////////////////

		if (!cmd_str.empty() && ch_is_white(cmd_str.back())) {
			con_xy(0+cursor_pos, CMD_Y);  // position cursor on cmd line
			vector<string> vtoken = str_to_tokens(cmd_str);
			int32_t tokens = vtoken.size();
			cmd_f = false;
			if (tokens) {
				// get last token
				string last_token = vtoken[tokens-1];
				// edit @ hilite
				if ((mode == edit) && hl_f) {
					cmd_f = true;
					err_f = !str_is_int(vtoken[0]);
					if (err_f) { cmd_msg = "in edit mode, expecting a number"; }
					if (!err_f) {
						int32_t ed_val = strtol(vtoken[0].c_str(), NULL, 0);  // STRTOL!
						if (ed_xy_on_title(ed_x, ed_y)) {  // on title: edit all
							ed_enc_wr(sp, ed_pre, ed_val, ed_x-1, ed_y-3, false);
							ed_enc_wr(sp, ed_pre, ed_val, ed_x-1, ed_y-2, false);
							ed_enc_wr(sp, ed_pre, ed_val, ed_x-1, ed_y-1, false);
							ed_enc_wr(sp, ed_pre, ed_val, ed_x-1, ed_y-0, false);
							ed_enc_wr(sp, ed_pre, ed_val, ed_x, ed_y-3, false);
							ed_enc_wr(sp, ed_pre, ed_val, ed_x, ed_y-2, false);
							ed_enc_wr(sp, ed_pre, ed_val, ed_x, ed_y-1, true);
						}
						else {  // edit single
							ed_val_wr(sp, ed_pre, ed_val, ed_x, ed_y, true);
						}
					}
				}
				// refresh screen
				else if ((last_token == "r") && (tokens == 1)) {
					cmd_f = true;
					err_f = false;
					dir_proc(dir_all, dir_dlp, false);  // update file list
					dir_idx = 0;
					hl_f = false; 
					con_cls();
				}
				// quit
				else if (((last_token == "quit") || (last_token == "exit")) && (tokens == 1)) {
					cmd_f = true;
					err_f = false;
					kbd_mode(false, true);  // restore orig keyboard mode
					sp_close_port(sp);  // close serial port
					#ifdef _MSWIN
						timeEndPeriod(1);
					#else
						kbd_mode(false, true);  // restore orig keyboard mode
					#endif					
					return(0);  // go bye-bye
				}
				// help mode
				else if ((last_token == "help") && (tokens == 1)) {
					cmd_f = true;
					err_f = view_file(help_fname, CON_H, false);  // call file viewer
					con_cls();  // clear screen
				}
				// terminal mode
				else if ((last_token == "term") && (tokens == 1)) {
					cmd_f = true;
					err_f = false;
					term(sp);  // call terminal sub
					con_cls();  // clear screen
				}
				// read SW version
				else if ((last_token == "ver") && (tokens == 1)) {
					cmd_f = true;
					err_f = version_rd(sp, cmd_msg);  // call sub
					if (err_f) { cmd_msg = "can't read SW version"; }
					con_cls();  // clear screen
				}
				// delete file
				else if (last_token == "delf") {
					cmd_f = true;
					err_f = (tokens != 2);
					if (err_f) { cmd_msg = "expecting 1 filename"; }
					if (!err_f) {
						string fnamex = file_ext_add(vtoken[0], DLP_EXT);
						err_f = file_delete(fnamex);
						if (err_f) { cmd_msg = "can't delete file:" + fnamex; }
					}
					dir_proc(dir_all, dir_dlp, false);  // update file list
					dir_idx = xy_move(dir_idx, 0, 0, dir_all.size()-1, DIR_ROWS);
					hl_f = false; 
					con_cls();  // clear screen
				}
				// rename | copy file
				else if ((last_token == "renf") || (last_token == "copf")) {
					cmd_f = true;
					string src_fnamex, dst_fnamex;
					err_f = (tokens != 3);
					if (err_f) { cmd_msg = "expecting 2 filenames"; }
					if (!err_f) {
						src_fnamex = file_ext_add(vtoken[0], DLP_EXT);
						err_f = !file_exists(src_fnamex);
						if (err_f) { cmd_msg = "source file doesn't exist:" + src_fnamex; }
					}
					if (!err_f) {
						dst_fnamex = file_ext_add(vtoken[1], DLP_EXT);
						err_f = file_exists(dst_fnamex);
						if (err_f) { cmd_msg = "target file exists:" + dst_fnamex; }
					}
					if (!err_f && (last_token == "renf")) {
						err_f = file_rename(src_fnamex, dst_fnamex);
						if (err_f) { cmd_msg = "can't rename file:" + src_fnamex; }
					}
					if (!err_f && (last_token == "copf")) {
						err_f = file_copy(src_fnamex, dst_fnamex);
						if (err_f) { cmd_msg = "can't copy file:" + src_fnamex; }
					}
					dir_proc(dir_all, dir_dlp, false);  // update file list
					dir_idx = xy_move(dir_idx, 0, 0, dir_all.size()-1, DIR_ROWS);
					hl_f = false; 
					con_cls();  // clear screen
				}
				// read user or system preset file to editor, optional input file name
				else if ((last_token == "ptoe") || (last_token == "ptoe_sys")) {
					cmd_f = true;
					string rd_fname, rd_fnamex;
					err_f = (tokens > 2);
					bool usr_f = (last_token == "ptoe");
					if (err_f) { cmd_msg = "expecting 0 or 1 filenames"; }
					if (!err_f) {
						rd_fname = (tokens == 2) ? vtoken[0] : pre_fname;
						rd_fnamex = file_ext_add(rd_fname, DLP_EXT);  // add extension
						err_f = dlp_to_ed(ed_pre, rd_fnamex, usr_f);
						if (err_f) { cmd_msg = "can't read file:" + rd_fnamex; }
					}
/*
					if (!err_f) { 
						err_f = ed_to_knobs(sp, ed_pre, usr_f);
						if (err_f) { cmd_msg = "can't upload knobs"; }
					}
*/
					if (!err_f) { 
						pre_fname = rd_fname; 
						mode = edit;  // edit mode
					}
					con_cls();  // clear screen
				}
				// write editor to user or system preset file, optional input file name
				else if ((last_token == "etop") || (last_token == "etop_sys")) {
					cmd_f = true;
					string wr_fname, wr_fnamex;
					err_f = (tokens > 2);
					if (err_f) { cmd_msg = "expecting 0 or 1 filenames"; }
					if (!err_f) {
						bool usr_f = (last_token == "etop");
						wr_fname = (tokens == 2) ? vtoken[0] : pre_fname;
						wr_fnamex = file_ext_add(wr_fname, DLP_EXT);  // add extension
						err_f = ed_to_dlp(ed_pre, wr_fnamex, usr_f);
						if (err_f) { cmd_msg = "can't write file:" + wr_fnamex; }
					}
					if (!err_f) { pre_fname = wr_fname; }
					dir_proc(dir_all, dir_dlp, false);  // update file list
					dir_idx = xy_move(dir_idx, 0, 0, dir_all.size()-1, DIR_ROWS);
					hl_f = false; 
					con_cls();  // clear screen
				}
				// print screen to text file
				else if (last_token == "print") {
					cmd_f = true;
					string wr_fname;
					err_f = (tokens > 2); 
					if (err_f) { cmd_msg = "expecting 0 or 1 filenames"; }
					if (!err_f) {
						wr_fname = prn_fname;  // safe default
						if (tokens == 2) { wr_fname = vtoken[0]; }
						if (mode == edit) {
							err_f = ed_print_to_file(ed_pre, wr_fname, pre_fname, true);  // append
						}
						else if (mode == files) {
							err_f = dir_print_to_file(wr_fname, dir_all, true);  // append
						}
						else if (mode == slots) {
							err_f = slots_print_to_file(wr_fname, dl_min, dl_names_sv, true);  // append
						}
						if (err_f) { cmd_msg = "can't write file:" + wr_fname; }
					}
					if (!err_f) { prn_fname = wr_fname; }
				}
				// rename editor preset
				else if (last_token == "rene") {
					cmd_f = true;
					err_f = (tokens != 2); 
					if (err_f) { cmd_msg = "expecting a filename"; }
					if (!err_f) { pre_fname = vtoken[0].c_str(); }
				}
				// knobs to editor
				else if ((last_token == "ktoe") && (tokens == 1)) {
					cmd_f = true;
					err_f = knobs_to_ed(sp, ed_pre, true);
					if (!err_f) { knobs_to_ed(sp, ed_pre, false); }
					if (err_f) { cmd_msg = "can't download from knobs"; }
					if (!err_f) { 
						mode = edit;  // edit mode
						hl_f = false; 
					}
				}
				// editor to knobs
				else if ((last_token == "etok") && (tokens == 1)) {
					cmd_f = true;
					err_f = ed_to_knobs(sp, ed_pre, true);
					if (err_f) { cmd_msg = "can't upload knobs"; }
				}
				// preset file to slot
				else if (last_token == "ptos") {
					cmd_f = true;
					int32_t slot = 0;
					string rd_fnamex;
					err_f = (tokens != 3); 
					if (err_f) { cmd_msg = "expecting file & slot"; }
					if (!err_f) {  // check slot limits
						if (!str_is_int(vtoken[1])) { err_f = true; }
						if (err_f) { cmd_msg = "expecting a number"; }
					}
					if (!err_f) {
						slot = strtol(vtoken[1].c_str(), NULL, 0);  // STRTOL!
						err_f = slot_error(slot);
						if (err_f) { cmd_msg = "slot out of range"; }
					}
					if (!err_f) {
						rd_fnamex = file_ext_add(vtoken[0], DLP_EXT);  // add extension
						err_f = !file_exists(rd_fnamex);
						if (err_f) { cmd_msg = "can't read file:" + rd_fnamex;}
					}
					if (!err_f) { 
						err_f = dlp_to_slot(sp, slot, rd_fnamex);
						if (err_f) { cmd_msg = "can't upload to slot:" + to_string(slot); }
					}
				}
				// slot to preset file
				else if (last_token == "stop") {
					cmd_f = true;
					int32_t slot = 0;
					string wr_fnamex;
					err_f = (tokens != 3); 
					if (err_f) { cmd_msg = "expecting slot & file"; }
					if (!err_f) {  // check slot limits
						err_f = !str_is_int(vtoken[0]);
						if (err_f) { cmd_msg = "expecting a number"; }
					}
					if (!err_f) {
						slot = strtol(vtoken[0].c_str(), NULL, 0);  // STRTOL!
						err_f = slot_error(slot);
						if (err_f) { cmd_msg = "slot out of range"; }
					}
					if (!err_f) { 
						wr_fnamex = file_ext_add(vtoken[1], DLP_EXT);  // add extension
						err_f = slot_to_dlp(sp, slot, wr_fnamex);
						if (err_f) { cmd_msg = "can't download from slot:" + to_string(slot); }
					}
					dir_proc(dir_all, dir_dlp, false);  // update file list
					dir_idx = xy_move(dir_idx, 0, 0, dir_all.size()-1, DIR_ROWS);
					hl_f = false; 
					con_cls();  // clear screen
				}
				// slots to preset numbered files
				else if (last_token == "stop_num") {
					cmd_f = true;
					err_f = (tokens != 3); 
					if (err_f) { cmd_msg = "expecting slot start & end"; }
					if (!err_f) {  // check slot
						err_f = !str_is_int(vtoken[0]);
						if (err_f) { cmd_msg = "expecting a number"; }
						if (!err_f) { dl_min = strtol(vtoken[0].c_str(), NULL, 0); }  // STRTOL!
					}
					if (!err_f) {  // check slot
						err_f = !str_is_int(vtoken[1]);
						if (err_f) { cmd_msg = "expecting a number"; }
						if (!err_f) { dl_max = strtol(vtoken[1].c_str(), NULL, 0); }  // STRTOL!
					}
					if (!err_f) {  // swap slots
						int32_t tmp_min = min(dl_min, dl_max);
						dl_max = max(dl_min, dl_max);
						dl_min = tmp_min;
					}
					if (!err_f) {  // check slot limits
						err_f = slot_error(dl_min);
						err_f |= slot_error(dl_max);
						if (err_f) { cmd_msg = "slot(s) out of range"; }
					}
					cmd_f = true;
					int32_t slot = dl_min;
					while ((!err_f) && (slot <= dl_max)) { 
						char buf[4];
						sprintf (buf, "%03d", slot);  // format to 3 decimals
						string slot_num(buf);						
						string wr_fnamex = file_ext_add(slot_num, DLP_EXT);  // add extension
						err_f = slot_to_dlp(sp, slot, wr_fnamex);
						if (err_f) { cmd_msg = "can't download from slot:" + slot_num; }
						if (!(slot % 4)) { cout << "." << flush; }  // show activity
						slot++;
					}
					dir_proc(dir_all, dir_dlp, false);  // update file list
					dir_idx = xy_move(dir_idx, 0, 0, dir_all.size()-1, DIR_ROWS);
					hl_f = false; 
					con_cls();  // clear screen
				}
				// editor to slot
				else if (last_token == "etos") {
					cmd_f = true;
					int32_t slot = 0;
					err_f = (tokens != 2); 
					if (err_f) { cmd_msg = "expecting slot"; }
					if (!err_f) {  // check slot limits
						if (!str_is_int(vtoken[0])) { err_f = true; }
						if (err_f) { cmd_msg = "expecting a number"; }
					}
					if (!err_f) {
						slot = strtol(vtoken[0].c_str(), NULL, 0);  // STRTOL!
						err_f = slot_error(slot);
						if (err_f) { cmd_msg = "slot out of range"; }
					}
					if (!err_f) { 
						err_f = ed_to_slot(sp, ed_pre, slot);
						if (err_f) { cmd_msg = "can't upload to slot:" + to_string(slot); }
					}
				}
				// slot to editor
				else if (last_token == "stoe") {
					cmd_f = true;
					int32_t slot = 0;
					err_f = (tokens != 2); 
					if (err_f) { cmd_msg = "expecting slot"; }
					if (!err_f) {  // check slot limits
						err_f = !str_is_int(vtoken[0]);
						if (err_f) { cmd_msg = "expecting a number"; }
					}
					if (!err_f) {
						slot = strtol(vtoken[0].c_str(), NULL, 0);  // STRTOL!
						err_f = slot_error(slot);
						if (err_f) { cmd_msg = "slot out of range"; }
					}
					if (!err_f) { 
						err_f = slot_to_ed(sp, ed_pre, slot);
						if (err_f) { cmd_msg = "can't download from slot:" + to_string(slot); }
					}
					if (!err_f) { 
						mode = edit;  // edit mode
					}
				}
				// slot(s) to list
				else if (last_token == "stol") {
					cmd_f = true;
					err_f = (tokens > 3);
					if (err_f) { cmd_msg = "expecting 0, 1, or 2 numbers"; }
					if (!err_f && (tokens > 1)) {  // 1 or 2 slots
						err_f = !str_is_int(vtoken[0]);
						if (err_f) { cmd_msg = "expecting a number"; }
						if (!err_f) { dl_min = strtol(vtoken[0].c_str(), NULL, 0); }  // STRTOL!
					}
					if (!err_f && (tokens == 3)) {  // 2 slots
						err_f = !str_is_int(vtoken[1]);
						if (err_f) { cmd_msg = "expecting a number"; }
						if (!err_f) { dl_max = strtol(vtoken[1].c_str(), NULL, 0); }  // STRTOL!
					}
					if (!err_f && (tokens == 2)) {  // 1 slot
						dl_max = dl_min;
					}
					if (!err_f && (tokens == 3)) {  // 2 slots
						int32_t tmp_min = min(dl_min, dl_max);
						dl_max = max(dl_min, dl_max);
						dl_min = tmp_min;
					}
					if (!err_f) {  // check slot limits
						err_f = slot_error(dl_min);
						err_f |= slot_error(dl_max);
						if (err_f) { cmd_msg = "slot(s) out of range"; }
					}
					if (!err_f) {  // dl slots & match contents with file names
						err_f = slots_to_pre_sv_names(sp, dl_min, dl_max, dir_dlp, dl_pre_sv, dl_names_sv);
						if (err_f) { cmd_msg = "can't download slot(s)"; }
					}
					if (!err_f) {
						dl_idx = 0;
						con_cls(); 
					}
				}
				// list to bank
				else if (last_token == "ltob") {
					cmd_f = true;
					string wr_fnamex;
					err_f = (tokens != 2); 
					if (err_f) { cmd_msg = "expecting file"; }
					if (!err_f) { 
						wr_fnamex = file_ext_add(vtoken[0], BNK_EXT);  // add extension
						err_f = sv_to_file(wr_fnamex, dl_names_sv, false);
					}
					dir_proc(dir_all, dir_dlp, false);  // update file list
					dir_idx = xy_move(dir_idx, 0, 0, dir_all.size()-1, DIR_ROWS);
					hl_f = false; 
					con_cls();  // clear screen
				}
				// bank to slots
				else if (last_token == "btos") {
					cmd_f = true;
					int32_t slot = 0;
					uint32_t idx = 0;
					string rd_fnamex;
					vector<string> bnk_sv;
					err_f = (tokens != 3); 
					if (err_f) { cmd_msg = "expecting file & slot"; }
					if (!err_f) {  // check slot input
						if (!str_is_int(vtoken[1])) { err_f = true; }
						if (err_f) { cmd_msg = "expecting a number"; }
					}
					if (!err_f) {  // check slot range
						slot = strtol(vtoken[1].c_str(), NULL, 0);  // STRTOL!
						err_f = slot_error_user(slot);
						if (err_f) { cmd_msg = "slot out of range"; }
					}
					if (!err_f) {  // read bank file
						rd_fnamex = file_ext_add(vtoken[0], BNK_EXT);  // add extension
						err_f = file_to_sv(rd_fnamex, bnk_sv);
						if (err_f) { cmd_msg = "can't read file:" + rd_fnamex;}
					}
					while ((!err_f) && (idx < bnk_sv.size()) && !slot_error_user(slot)) { 
						if (bnk_sv[idx].find("-?-") == string::npos) {  // if not "-?-"
							string dlp_fnamex = file_ext_add(bnk_sv[idx], DLP_EXT);  // add extension
							err_f = dlp_to_slot(sp, slot, dlp_fnamex);
							if (err_f) { cmd_msg = "can't upload file:" + dlp_fnamex + " to slot:" + to_string(slot); }
						}
						idx++;
						if (!(idx % 4)) { cout << "." << flush; }  // show activity
						if (slot < 0) { slot--; } // go neg if starting neg
						else { slot++; }
					}
				}
				// spi pump
				else if (last_token == "spi_pump") {
					cmd_f = true;
					err_f = (tokens != 2); 
					if (err_f) { cmd_msg = "expecting a filename"; }
					if (!err_f) {
						err_f = spi_pump(sp, vtoken[0]);
						if (err_f) { cmd_msg = "can't pump spi file:" + vtoken[0]; }
					}
				}
				// change directory
				else if ((last_token == "cd") || (last_token == "chdir")) {
					cmd_f = true;
					err_f = (tokens != 2);
					if (err_f) { cmd_msg = "expecting directory name"; }
					if (!err_f) {
						err_f = chdir(vtoken[0].c_str());
						if (err_f) { cmd_msg = "can't open directory:" + vtoken[0]; }
					}
					dir_proc(dir_all, dir_dlp, false);  // update file list
					dir_idx = 0;
					hl_f = false; 
					con_cls();  // clear screen
				}
				// dir / ls
				else if (((last_token == "dir") || (last_token == "ls")) && (tokens == 1)) {
					cmd_f = true;
					err_f = false;
					dir_proc(dir_all, dir_dlp, true);  // update file list
					dir_idx = 0;
					hl_f = false; 
					con_cls();  // clear screen
				}
				// update preset file
				else if (last_token == "up") {
					cmd_f = true;
					err_f = (tokens != 2);
					if (err_f) { cmd_msg = "expecting 1 filename"; }
					if (!err_f) {
						string fnamex = file_ext_add(vtoken[0], DLP_EXT);
						err_f = dlp_update(fnamex);
						if (err_f) { cmd_msg = "can't update file:" + fnamex; }
					}
				}
				// do command line stuff
				if (cmd_f) { 
					cmd_prev = cmd_str;  // keep old
					cmd_hist.wr(cmd_str);  // store & clear cmd
					cursor_pos = cmd_str.size();  // reset cursor
					if (err_f) { cmd_msg = "<?> " + cmd_msg; }  
					else { cmd_msg = "<OK> " + cmd_msg; }
					cmd_prev += cmd_msg;  // keep old
					cmd_msg = "";  // clear
				}
			}
		}

		//////////
		// draw //
		//////////

		// editor view
		if (mode == edit) {
			con_xy(0, 0);
			ed_draw(ed_pre, ed_x, ed_y, hl_f);
		}

		// directory view
		if (mode == files) {
			con_xy(0, 0);
			dir_draw(dir_idx, hl_f, dir_all);
		}

		// slots view
		if (mode == slots) {
			con_xy(0, 0);
			slots_draw(dl_idx, hl_f, dl_min, dl_names_sv);
		}

		// draw preset file name
		con_font('b');  // bold
		con_xy(0, FILE_Y);
		cout << setw(CMD_W - 40) << left << ("preset: " + pre_fname);
		cout << setw(40) << right << ("print: " + prn_fname);
		cout << endl << setw(CMD_W) << "";

		// draw cmd line
		if (err_f) { con_font('e'); }  // hilite errors
		else { con_font('n'); }  // normal
		con_xy(0, PREV_Y); cout << setw(CMD_W) << left << cmd_prev;
		con_font('d');  // default
		con_xy(0, CMD_Y); cout << setw(CMD_W) << left << cmd_str;
		con_xy(0+cursor_pos, CMD_Y);  // position cursor on cmd line

		// clear flag
		first_f = false;

	}


} // end of main

